import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.nio.ByteBuffer;
import java.util.Arrays;

import mp3tools.Frame;
import mp3tools.Header;
import mp3tools.IllegalHeaderException;
import mp3tools.MP3Buffer;



public class mp3extract {

	
	private static int sizeHead = 0; //Size of sample headers.
	public static void main(String[] args)
	{
		try
		{
		if (args.length != 1)
		{
			printUsage();
			return;
		}
		if (!args[0].endsWith(".fsb"))
		{
			System.out.println("Song File must be in fsb format!");
			printUsage();
			return;
		}
		File mp3File = new File(args[0]);
		InputStream mp3Input = new FileInputStream(mp3File);
		MP3Buffer mp3Buffer = new MP3Buffer((int) mp3File.length());
		System.out.println("Reading FSB file... "+mp3File.length()+" bytes");
		mp3Input.read(mp3Buffer.array());
		//Check to see if this is actually fsb 4 file...
		byte[] headerCheck = {70, 83, 66, 52};
		byte[] header = new byte[4];
		System.arraycopy(mp3Buffer.array(), 0, header, 0, 4);
		if (!Arrays.equals(headerCheck, header))
		{
			System.out.println("Unknown header: Not a FSB4 file.");
			printUsage();
			return;
		}
		mp3Input.close();
		//Get filename from fsb4 file.
		String fileName = "";
		for (int i=0; i<30; i++)
		{
			fileName = fileName+((char) mp3Buffer.array()[50+i]);
		}
		//Now format white space out of string.
		fileName = fileName.trim();
		System.out.println("Formatting filename...");
		int numChannels = 0;
		for (int i=1; i>-1; i--)
		{
			int numadd = mp3Buffer.array()[110+i];
			if (numadd<0) numadd += 256; //WE need to convert byte to int. Since java uses signed bytes
			//we need to add 256 to negative numbers. I guess we could unary minus it (flip bits add 1)
			//But using decimal math rather than binary math is easier.
			numChannels += numadd*((int) (Math.pow(16, i*2)));
		}
		for (int i=3; i>-1; i--)
		{
			int numadd = mp3Buffer.array()[8+i];
			if (numadd<0) numadd += 256; //WE need to convert byte to int. Since java uses signed bytes
			//we need to add 256 to negative numbers. I guess we could unary minus it (flip bits add 1)
			//But using human math rather than binary math is easier.
			sizeHead += numadd*((int) (Math.pow(16, i*2)));
		}
		//Index is at sizeHeaders-1+4.
		//numChannels is represented in mono, so we have to divide by two as all ghwt songs are
		//stereo.
		//if mono, we don't divide by 2. First frame is decided... mono/vs stereo
		//Get size of sample header.
		if (((mp3Buffer.array()[48+sizeHead+3] >>> 6) & 3) == 3)
			//Byte order 11 is mono!
		{
			//NumChannels stays the same.
		}
		else
			numChannels /=2;
		//MAIN EXTRACTION CODE!!
		for (int i=0; i<numChannels; i++)
		{
			System.out.println("Extracting "+fileName+" channel "+(i+1)+"...");
			//int lastIndexOfSlash = args[0].lastIndexOf("\\");
			//if (lastIndexOfSlash == -1) lastIndexOfSlash = 0; //No slashes in filename!
			String outputname = mp3File.getAbsolutePath().substring(0, (int) (mp3File.getAbsolutePath().length() - mp3File.getName().length())).concat(fileName).concat("_channel_"+(i+1)+".mp3");
			OutputStream channelOutput = new FileOutputStream(outputname);
			ByteBuffer bb = extractChannel(mp3Buffer, i, numChannels);
			System.out.println("Writing "+fileName+" channel "+(i+1)+"...");
			channelOutput.write(bb.array());
			channelOutput.close();
		}
	 }
	 catch (IOException e)
	 {
		 System.err.println("Caught IO Exception: cannot read/write data!: " + e.getMessage());
	 }
	}
	
	private static void printUsage()
	{
		System.out.println("Usage.....");
		System.out.println("java mp3extract <song.fsb> ");
		System.out.println("");
		System.out.println("Extracts individual channels from fsb file.");
	}
	
	private static ByteBuffer extractChannel(MP3Buffer mp3Buffer, int curChannel, int numChannels) throws IOException
	{
		//ByteArrayOuputStream is used because bytebuffer backs a constant array, while
		//It is possible to change size of outputstream dynamically, and produce
		//a byte array for exporting.
		ByteArrayOutputStream channelStream = new ByteArrayOutputStream();
		boolean eofReached = false;
		//Start of Mp3 Data.
		mp3Buffer.position(48+sizeHead);
		//Start w/ 0th Channel.
		int curChannelView = 0;
		while(!eofReached)
		{
			//Read every frame's bunch of bits. Determine frame length.
			//Error checking routines are now covered by the mp3tools package!!
			Frame nextFrame = null;
			try {
				nextFrame = mp3Buffer.getNextFrame();
				if (!((nextFrame.getHeader().getFrameLength() % 16) == 0))
				{
					//update buffer to reflect fsb byte-alignment.
					mp3Buffer.position(mp3Buffer.position()-nextFrame.getHeader().getFrameLength()+((nextFrame.getHeader().getFrameLength() / 16)+1)*(16));
				}
			} catch (IllegalHeaderException e) {
                break;
			}
			if (curChannelView == curChannel)
			{
				byte[] frame = nextFrame.getFrameData();
				channelStream.write(frame);
			}
			//Otherwise, we just want to skip it.
			else
			{
				//Do nothing, position is already updated by call to MP3Buffer.getNextFrame()
			}
			curChannelView++;
			//Reset channel num.
			if (curChannelView == numChannels)
			{
				curChannelView = 0;
			}
			//Check if EOF is reached!
			if (!mp3Buffer.hasRemaining())
			{
				eofReached = true;
			}
		}
		return ByteBuffer.wrap(channelStream.toByteArray());
	}
}

